from loguru import logger

import DB
from .Speakers import ModelDescription
from .TTSServer import TTSServer


class TTSServerCached(TTSServer):
    def __init__(self):
        super().__init__()
        self.models_lookup_idx: dict[str, ModelDescription] = dict()

    async def synthesize_text(self, text: str, speaker: ModelDescription) -> bytes:
        cache_query = DB.SoundCache.select() \
            .where(DB.SoundCache.text == text) \
            .where(DB.SoundCache.speaker == speaker)

        if cache_query.count() == 1:
            with DB.database.atomic():
                DB.SoundCache.update({DB.SoundCache.usages: DB.SoundCache.usages + 1}) \
                    .where(DB.SoundCache.text == text) \
                    .where(DB.SoundCache.speaker == speaker).execute()

                cached = cache_query.get().audio

            return cached

        else:
            synthesized = await super().synthesize_text(text, speaker)
            DB.SoundCache.create(text=text, speaker=speaker, audio=synthesized)
            return synthesized

    async def discovery(self) -> list[ModelDescription]:
        res = await super().discovery()
        logger.debug(f'Discovered {len(res)} models')
        self.models_lookup_idx: dict[str, ModelDescription] = {f'{desc.engine}_{desc.name}': desc for desc in res}
        return res

    async def speaker_to_description(self, speaker_str: str) -> ModelDescription | None:
        for attempt in range(2):
            if len(self.models_lookup_idx) > 0:
                break

            await self.discovery()

        else:
            raise RuntimeError('Models discovery seems to return zero models')

        # Default to first model
        return self.models_lookup_idx.get(speaker_str, None)

    def default_speaker(self) -> ModelDescription:
        return list(self.models_lookup_idx.values())[0]
